/**
  *  \file arch/win32/console.cc
  *  \brief Console Implementation for Windows
  *
  *  Windows has distinct calls for accessing consoles using a
  *  determined character set ("ANSI" codepage or Unicode). We
  *  therefore check whether we're actually talking to a console and
  *  convert to the appropriate character set.
  *
  *  When accessing files, we assume UTF-8. This has the advantage of
  *  being able to receive anything we produce, without requiring the
  *  user to configure an output codepage. Using the console codepage
  *  would not be a good idea, because that one usually differs from
  *  the GUI codepage, making the output look garbled in usual
  *  editors. In contrast, Notepad (and surely most other decent
  *  Windows editors) are capable of UTF-8 and will auto-detect it.
  *
  *  For reading, we display prompts only if both stdin and stdout
  *  actually are consoles. This way, prompts don't pile up on stdout
  *  when stdin is redirected. The disadvantage is that we don't get
  *  prompts when stdin is Emacs' pipe (but that one's different
  *  anyway because it expects UTF-8, like all non-consoles). In
  *  addition, it seems that we have to do our own Ctrl-C processing
  *  for consoles(?).
  */

#include <windows.h>
#include "arch/win32-bcc55/platform.h"
#include "cpluslib/utf8.h"
#include "arch/win32/win32.h"
#include "util/console.h"
#include "io/streams.h"

namespace {
    struct ConsoleStream {
        enum Kind {
            isNullStream,
            isConsole,
            isFile
        };
        HANDLE handle;
        Kind kind;

        void init(DWORD what);
    };
}

void
ConsoleStream::init(DWORD what)
{
    DWORD tmp;
    handle = GetStdHandle(what);
    if (handle == 0 || handle == INVALID_HANDLE_VALUE) {
        // Happens with GUI apps
        kind == isNullStream;
    } else if (GetConsoleMode(handle, &tmp)) {
        // It's a console
        kind = isConsole;
    } else {
        // It's a file
        kind = isFile;
    }
}

static ConsoleStream csIn, csOut, csErr;

/** Windows Console Handler, Unicode version (for Windows NT). */
static void
consoleHandlerW(const Console& con, int n)
{
    while (n--) {
        const Console::Message& m = con.getLine(n);
        ConsoleStream& stream = (m.mc >= LOG_WARN ? csErr : csOut);
        if (stream.kind == ConsoleStream::isConsole) {
            // Write to console
            DWORD written;
            win32::WStr wstr;
            win32::convertToUnicode(wstr, m.text, false);
            wstr.push_back(L'\r');
            wstr.push_back(L'\n');
            WriteConsoleW(stream.handle, &wstr[0], wstr.size(), &written, 0);
        } else if (stream.kind == ConsoleStream::isFile) {
            // FIXME: can we do any sensible error handling here?
            DWORD written;
            if (m.text.size())
                WriteFile(stream.handle, m.text.data(), m.text.size(), &written, 0);
            WriteFile(stream.handle, "\r\n", 2, &written, 0);
        } else {
            // it's redirected, so don't even bother
        }
    }
}

/** Windows Console Handler, ANSI version (for Windows 95). */
static void
consoleHandlerA(const Console& con, int n)
{
    while (n--) {
        const Console::Message& m = con.getLine(n);
        ConsoleStream& stream = (m.mc >= LOG_WARN ? csErr : csOut);
        if (stream.kind == ConsoleStream::isConsole) {
            // Write to console
            DWORD written;
            string_t str(win32::convertToCodepage(m.text.c_str(), GetConsoleOutputCP()));
            str += "\r";
            str += "\n";
            WriteConsoleA(stream.handle, str.data(), str.size(), &written, 0);
        } else if (stream.kind == ConsoleStream::isFile) {
            // FIXME: can we do any sensible error handling here?
            DWORD written;
            if (m.text.size())
                WriteFile(stream.handle, m.text.data(), m.text.size(), &written, 0);
            WriteFile(stream.handle, "\r\n", 2, &written, 0);
        } else {
            // it's redirected, so don't even bother
        }
    }
}

void
Platform::initConsoleOutput()
{
    csIn.init(STD_INPUT_HANDLE);
    csOut.init(STD_OUTPUT_HANDLE);
    csErr.init(STD_ERROR_HANDLE);

    if (win32::hasUnicodeSupport()) {
        console.console_changed.add(consoleHandlerW);
    } else {
        console.console_changed.add(consoleHandlerA);
    }
}

namespace {
    struct ConsoleReader {
        union {
            char charBuffer[1024];
            wchar_t wcharBuffer[512];
        } buffer;
        std::size_t fill;
        std::size_t ptr;

        bool readLineA(string_t& out, bool fromConsole);
        bool readLineW(string_t& out);
    };
}

static ConsoleReader reader;  /* rely on zero-initialisation */

/** Read a line using byte syscalls. Produces the output in the
    respective byte representation.
    \param out [out] line; input will be appended
    \param fromConsole [in] true to use ReadConsoleA and produce console codepage, false to use ReadFile
    \return true iff line read, false on EOF/error */
bool
ConsoleReader::readLineA(string_t& out, bool fromConsole)
{
    while (1) {
        if (ptr >= fill) {
            DWORD read;
            BOOL result = (fromConsole
                           ? ReadConsoleA(csIn.handle, buffer.charBuffer, sizeof(buffer.charBuffer), &read, 0)
                           : ReadFile    (csIn.handle, buffer.charBuffer, sizeof(buffer.charBuffer), &read, 0));
            if (result && read != 0) {
                ptr = 0;
                fill = read;
            } else {
                return out.size() != 0;
            }
        }

        char c = buffer.charBuffer[ptr++];
        if (c == '\r') {
            // ignore
        } else if (c == '\n') {
            // finish
            return true;
        } else if (fromConsole && c == 26) {
            // Ctrl-Z
            --ptr;
            return out.size() != 0;
        } else {
            out += c;
        }
    }
}

/** Read a line using Unicode syscalls. Produces the output in UTF-8.
    \param out [out] line; input will be appended
    \return true iff line read, false on EOF/error */
bool
ConsoleReader::readLineW(string_t& out)
{
    while (1) {
        if (ptr >= fill) {
            DWORD read;
            BOOL result = ReadConsoleW(csIn.handle, buffer.wcharBuffer, sizeof(buffer.wcharBuffer) / sizeof(buffer.wcharBuffer[0]), &read, 0);
            if (result && read != 0) {
                ptr = 0;
                fill = read;
            } else {
                return out.size() != 0;
            }
        }

        wchar_t c = buffer.wcharBuffer[ptr++];
        if (c == L'\r') {
            // ignore
        } else if (c == L'\n') {
            // finish
            return true;
        } else if (c == 26) {
            // Ctrl-Z
            --ptr;
            return out.size() != 0;
        } else {
            appendUtf8(out, c);
        }
    }
}

/** Read console line.
    \param prompt [in] Prompt to display before reading.
    \param line [out] Line is produced here (not appended)
    \return true on success, false on error/EOF */
bool
Platform::readConsoleLine(string_t prompt, string_t& line)
{
    if (csIn.kind == ConsoleStream::isFile) {
        line.clear();
        return reader.readLineA(line, false);
    } else if (csIn.kind == ConsoleStream::isConsole) {
        if (win32::hasUnicodeSupport()) {
            // Display prompt when we are interactive.
            if (csOut.kind == ConsoleStream::isConsole) {
                win32::WStr wstr;
                DWORD written;
                win32::convertToUnicode(wstr, prompt, false);
                WriteConsoleW(csOut.handle, &wstr[0], wstr.size(), &written, 0);
            }

            // Now read user input
            line.clear();
            return reader.readLineW(line);
        } else {
            // Display prompt when we are interactive.
            if (csOut.kind == ConsoleStream::isConsole) {
                string_t promptA(win32::convertToCodepage(prompt.c_str(), GetConsoleOutputCP()));
                DWORD written;
                WriteConsoleA(csOut.handle, promptA.data(), promptA.size(), &written, 0);
            }

            // Now read user input
            string_t tmp;
            if (reader.readLineA(tmp, true)) {
                line = win32::convertFromCodepage(tmp.c_str(), GetConsoleCP());  /* FIXME: .c_str() */
                return true;
            } else {
                return false;
            }
        }
    } else {
        return false;
    }
}
