/**
  *  \file client/cm-select.cc
  *  \brief Starchart Mode: Range Selection
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/cm-select.h"
#include "game/sel.h"
#include "gfx/palette.h"
#include "gfx/gfxcontext.h"
#include "game/global.h"
#include "client/chartloc.h"
#include "client/chart.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/obj.h"
#include "game/ship.h"
#include "game/planet.h"
#include "cpluslib/format.h"

/** Constructor.
    \param origin Point of invocation */
WSelectChartMode::WSelectChartMode(GPoint origin)
    : WSimpleChartMode(ChartMode_PrimaryMode),
      origin(origin), end(origin),
      nobjs(0)
{
    // Store current selection
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    GMultiSelection::getSelectionLayer(GMultiSelection::getCurrentSelectionLayer()).copyFrom(univ);
}

WSelectChartMode::~WSelectChartMode()
{ }

/** Check and mark an object. If the object is in range, it will be marked.
    \param obj Object to check */
void
WSelectChartMode::checkObject(GMapObject& obj)
{
    if (obj.isExistant()) {
        // Bounding rectangle in correct size
        int ax = std::min(origin.x, end.x);
        int bx = std::max(origin.x, end.x);
        int ay = std::min(origin.y, end.y);
        int by = std::max(origin.y, end.y);

        // Location. Try to move into bounding rectangle if it's outside.
        GPoint pt = obj.getPos();
        if (chart_conf.wrapped) {
            if (pt.x < ax) {
                pt.x += chart_conf.xsize;
            }
            if (pt.x > bx) {
                pt.x -= chart_conf.xsize;
            }
            if (pt.y < ay) {
                pt.y += chart_conf.ysize;
            }
            if (pt.y > by) {
                pt.y -= chart_conf.ysize;
            }
        }

        // Inside?
        if (pt.x >= ax && pt.x <= bx && pt.y >= ay && pt.y <= by) {
            ++nobjs;
            obj.setIsMarked(true);
        }
    }
}

/** Rebuild selection from scratch. */
void
WSelectChartMode::rebuildSelection()
{
    // Clear status, reload selection from multi-selection buffer
    nobjs = 0;
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    GMultiSelection::getSelectionLayer(GMultiSelection::getCurrentSelectionLayer()).copyTo(univ);

    // Update
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        checkObject(univ.getShip(sid));
    }
    for (int pid = 1; pid <= NUM_PLANETS; ++pid) {
        checkObject(univ.getPlanet(pid));
    }

    univ.doScreenUpdates();
}

/** Handle movement. This updates the origin/end points. */
void
WSelectChartMode::onMove(GChartLocation& loc)
{
    origin += loc.getWrapAdjust();
    if (end != loc.getLocation()) {
        end = loc.getLocation();
        loc.markChanged();
        rebuildSelection();
    }
    WSimpleChartMode::onMove(loc);
}

/** Draw below. This draws the selection frame. */
void
WSelectChartMode::drawBelow(GfxCanvas& can, GChartLocation& loc)
{
    WSimpleChartMode::drawOverlays(can, loc);

    // Build context
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_SHIELD + 4);

    // Figure screen points for drawing
    GChartViewport& vp = loc.getViewport();
    GfxPoint originScaled = vp.scaleXY(origin);
    GfxPoint endScaled    = vp.scaleXY(end);

    if (endScaled == originScaled) {
        // Single point
        drawPixel(ctx, originScaled);
    } else {
        // Area
        GfxRect r(originScaled, 1, 1);
        r.include(endScaled);
        drawRectangle(ctx, r);

        // Content of area if present
        if (r.w > 2 && r.h > 2) {
            r.grow(-1, -1);
            drawSolidBar(ctx, r, COLOR_SHIELD + 3);
        }
    }
}

/** Draw overlays. Just a simple helpful text. */
void
WSelectChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    GfxRect pos = loc.getViewport().view;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_WHITE);
    ctx.useFont(*fonts[FONT_NORMAL]);
    outText(ctx, pos.x, pos.y, _("Mark range, end with [ESC], cancel with [Backspace]."));
    if (nobjs != 0) {
        outText(ctx, pos.x, pos.y + font_heights[FONT_NORMAL], format(_("%d unit%!1{s%} in range."), nobjs));
    }

    WSimpleChartMode::drawOverlays(can, loc);
}

bool
WSelectChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    // FIXME: what about selection switching? PCC 1.x handles this explicitly
    // for Alt-Left/Right, but doesn't handle Alt-. or '.' specially.
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_ESCAPE:
         case 'r':
            // Exit mode
            delete this;
            return true;

         case SDLK_BACKSPACE:
            // Cancel
            {
                // Restore from buffer
                GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
                GMultiSelection::getSelectionLayer(GMultiSelection::getCurrentSelectionLayer()).copyTo(univ);
                univ.doScreenUpdates();

                // Back to origin and exit
                loc.setLocation(origin);
                delete this;
            }
            return true;
        }
    }
    return WSimpleChartMode::handleEvent(event, loc);
}
