/**
  *  \file client/dialogs/combatdlg.cc
  *  \brief Combat Overview Dialog
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/combatdlg.h"
#include "client/widgets/combatdiagram.h"
#include "client/widgets/racelist.h"
#include "game/vcrdb.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widget-busy.h"
#include "ui/widgets/button.h"
#include "ui/widgets/keytranslator.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/progress.h"
#include "ui/widgets/checkbox.h"

namespace {
    class WCombatDiagramDialog : public UIWindow {
        ObjectHolder h;
        WCombatDiagramWidget* diagWidget;
        UICheckbox* colorCheckbox;
        int chosenBattle;

        void init(GClassicVcrDatabase& db, WCombatDiagram& diag);
        void onClickBattle();
        void onToggleTeamColors();
     public:
        WCombatDiagramDialog(GClassicVcrDatabase& db, WCombatDiagram& diag);
        int getChosenBattle() const
            { return chosenBattle; }
    };
}

WCombatDiagramDialog::WCombatDiagramDialog(GClassicVcrDatabase& db, WCombatDiagram& diag)
    : UIWindow(_("Combat Overview"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      chosenBattle(-1)
{
    init(db, diag);
}

void
WCombatDiagramDialog::init(GClassicVcrDatabase& db, WCombatDiagram& diag)
{
    // VBox
    //   WCombatDiagramWidget
    //   WRaceList
    //   HBox
    //     UIButton "OK"
    //     UICheckbox "Team colors"
    //     UISpacer
    setLayout(UIVBoxLayout::instance5);

    diagWidget = &h.add(new WCombatDiagramWidget(db, diag));
    diagWidget->sig_click_battle.add(this, &WCombatDiagramDialog::onClickBattle);
    add(*diagWidget);
    add(h.add(new WRaceList(30*font_ems[FONT_NORMAL], WRaceList::FlowLayout | WRaceList::ShowRaceNames, 1, NUM_OWNERS)));

    colorCheckbox = &h.add(new UICheckbox(0, _("Team colors"), 't', 2));
    colorCheckbox->setValue(diagWidget->getTeamColorState());
    colorCheckbox->sig_changed.add(this, &WCombatDiagramDialog::onToggleTeamColors);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UIButton(1, _("Close"), SDLK_RETURN)).addStop());
    g.add(*colorCheckbox);
    g.add(h.add(new UISpacer()));
    add(g);
    add(h.add(new UIKeyTranslator()).addStop(' ', 1).addStop(SDLK_ESCAPE, 1));

    GfxRect pos = ui_root->getExtent();
    pos.grow(-10, -10);
    setExtent(pos);
}

void
WCombatDiagramDialog::onClickBattle()
{
    // Must save the chosen battle because closing the window will
    // invalidate it inside the diagWidget.
    chosenBattle = diagWidget->getHoverSlot();
    stop(2);
}

void
WCombatDiagramDialog::onToggleTeamColors()
{
    diagWidget->setTeamColorState(colorCheckbox->getValue());
}

/** Combat overview dialog. Displays a WCombatDiagramWidget.
    \param db VCRs to display
    \return clicked VCR number [0,db.getCount()), if any; otherwise -1 */
int
doCombatDialog(GClassicVcrDatabase& db)
{
    // Need at least two VCRs
    if (db.getCount() < 2)
        return -1;

    // Initialize
    WCombatDiagram diag;
    {
        ProgressMonitor progress;
        UIBusyWidget busy;
        ui_root->add(busy);
        ui_root->performDeferredRedraws();
        progress.sig_tick.add(&busy, &UIBusyWidget::tick);

        diag.init(db, progress);
    }

    // Display dialog
    WCombatDiagramDialog win(db, diag);
    if (ui_root->runModal(win) == 2) {
        return win.getChosenBattle();
    } else {
        return -1;
    }
}
