/**
  *  \file client/wscorechart.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wscore.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "game/stat.h"
#include "util/nls.h"
#include "gfx/font.h"
#include "gfx/palette.h"
#include "game/host.h"
#include "game/config.h"
#include "gfx/clip.h"
#include "game/team.h"
#include "client/widgets/racelist.h"
#include "ui/widgets/tooltip.h"
#include "cpluslib/format.h"

/** Constructor.
    \param file  Score statistics file
    \param first First turn index to use
    \param last  Last turn index to use */
WScoreChart::WScoreChart(const GStatFile& file, uint32 first, uint32 last)
    : file(file),
      first(first),
      last(last),
      score("?"),
      current_player(0),
      aggreg(ScoreByPlayer),
      left_margin(0),
      right_margin(0),
      bottom_margin(0),
      max_score(0),
      victory_level(-1),
      victory_turns_needed(0),
      tt_ctl(*this)
{
    victory_turns.setAll(0);
    recomputeGrid();
}

/** Set score to display. */
void
WScoreChart::setScore(WScore score)
{
    this->score = score;
    recomputeGrid();
    drawWidget();
}

/** Set aggregation mode. */
void
WScoreChart::setAggregation(WScoreAggregation aggr)
{
    if (aggr != this->aggreg) {
        this->aggreg = aggr;
        recomputeGrid();
        drawWidget();
    }
}

/** Set current player. This player will be highlighted using a thicker line. */
void
WScoreChart::setCurrentPlayer(int player)
{
    if (current_player != player) {
        current_player = player;
        drawWidget();
    }
}

void
WScoreChart::getLayoutInfo(LayoutInfo& info)
{
    info.min_size  = GfxPoint(left_margin + right_margin + 200, 5*bottom_margin);
    info.pref_size = GfxPoint(left_margin + right_margin + 400, 10*bottom_margin);
    info.flags     = li_Flexible;
}

void
WScoreChart::drawContent(GfxCanvas& can)
{
    /* can we draw? */
    if (extent.w < left_margin + right_margin + 2 || extent.h < bottom_margin)
        return;

    /* figure out position of grid */
    const int x1 = extent.x + left_margin;
    const int x2 = extent.x + extent.w - right_margin;
    const int y1 = extent.y + 1;
    const int y2 = extent.y + extent.h - bottom_margin;
    const int height = extent.h - bottom_margin - 1;
    const int width  = extent.w - right_margin - left_margin - 2;

    /* draw background */
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, extent);

    /* auxiliary grid */
    ctx.setColor(UIColor::tc_Faded);
    ctx.setLinePattern(DOTTED_LINE);
    ctx.useFont(*fonts[FONT_NORMAL]);
    if (aux_decay == 0) {
        /* no decay - simple lines */
        int32 value = aux_scale;
        while (value <= aux_limit) {
            int y = y2 - height*value / max_score;
            drawHLine(ctx, x1+1, y, x2);
            ctx.setTextAlign(2, 1);
            outText(ctx, x1-2, y, itoa(value));
            value += aux_scale;
        }
    } else {
        /* decay lines */
        enum { MAX_LINES = 20 };
        int32 ys[MAX_LINES];
        unsigned int nlines = 0;
        int lost = 0;

        GfxClipRectFilter clip(GfxRect(x1, y1, width, height), can);
        ctx.useCanvas(clip);

        for (int32 v = aux_scale; v <= aux_limit && nlines < MAX_LINES; v += aux_scale)
            ys[nlines++] = v;

        if (nlines >= 3) {
            /* We have sensible starting positions. When scores are too small, we'd end
               up with no sensible lines here, so draw no grid at all. */
            const int turns = file.getRecordByIndex(last-1).getTurnNumber() - file.getRecordByIndex(first).getTurnNumber();
            for (int i = 0; i < turns; ++i) {
                /* Compute x's */
                int xold = x1 + (width * i) / turns;
                int xnew = x1 + (width * (i+1)) / turns;

                /* Draw each line and adjust ys[] */
                for (unsigned int t = 0; t < nlines; ++t) {
                    int32 vold = ys[t];
                    int32 vnew = (ys[t] * (100 - aux_decay) + 50) / 100;
                    int32 yold = y2 - height*vold / max_score;
                    int32 ynew = y2 - height*vnew / max_score;

                    /* Alternate between 0xAA and 0x55 line styles to avoid generating
                       stripes at the bottom, where all lines run together. Note that
                       when we lose lines below, we still keep existing lines' styles,
                       so we have to keep track of how many we lost. */
                    ctx.setLinePattern(0xAA >> ((t+lost) & 1));
                    drawLine(ctx, GfxPoint(xold, yold), GfxPoint(xnew, ynew));

                    ys[t] = vnew;
                }

                /* When our highest line drops below the limit, add a new one. */
                while (ys[nlines-1] < aux_limit) {
                    if (nlines >= MAX_LINES) {
                        /* No room to add it, lose one near the y axis. */
                        for (unsigned i = 1; i < nlines; ++i)
                            ys[i-1] = ys[i];
                        --nlines, ++lost;
                    }
                    ys[nlines] = ys[nlines-1] + aux_scale;
                    ++nlines;
                }
            }
        }
    }
    ctx.useCanvas(can);
    ctx.setColor(UIColor::tc_Static);
    ctx.setLinePattern(SOLID_LINE);

    /* draw grid */
    drawVLine(ctx, x1, y1, y2);       /* Y (score) axis */
    drawHLine(ctx, x1, y2, x2);       /* X (time) axis */
    drawHLine(ctx, x1-1, y1+1, x1+1); /* Y arrowhead */
    drawHLine(ctx, x1-2, y1+2, x1+2);
    drawVLine(ctx, x2-1, y2-1, y2+1); /* X arrowhead */
    drawVLine(ctx, x2-2, y2-2, y2+2);

    ctx.setTextAlign(2, 0);
    outText(ctx, x1-2, y1, numToString(max_score));
    outText(ctx, x1-2, y2, numToString(file.getRecordByIndex(first).getTurnNumber()));
    ctx.setTextAlign(0, 0);
    if (isCumulative(aggreg)) {
        outText(ctx, x1+2, y1, _("Cumulative chart"));
    }
    outText(ctx, x2+2, y2, numToString(file.getRecordByIndex(last-1).getTurnNumber()));
    ctx.setTextAlign(0, 2);
    outText(ctx, x2+2, y2, _("Turn"));
    ctx.setTextAlign(2, 2);
    outText(ctx, x1-2, y2, _("Turn"));

    /* Win limit */
    if (victory_level > 0) {
        ctx.setLinePattern(DOTTED_LINE);

        int height = extent.h - bottom_margin;
        int y      = y2 - height*victory_level / max_score;
        drawHLine(ctx, x1, y, x2);
    }

    /* draw chart */
    ctx.setLineThickness(1);
    ctx.setLinePattern(SOLID_LINE);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (i != current_player && present_curves.contains(i)) {
            drawPlayerChart(ctx, i);
        }
    }
    if (current_player >= 1 && current_player <= NUM_PLAYERS) {
        ctx.setLineThickness(3);
        drawPlayerChart(ctx, current_player);
    }

    can.update(extent);
}

/** Recompute all measurement information. */
void
WScoreChart::recomputeGrid()
{
    /* Compute victory information */
    findVictoryCondition();

    /* Figure out maximum score */
    max_score = score.getMax(file, first, last, aggreg);
    if (max_score <= 0)
        max_score = 1;

    /* Figure out presence of curves */
    present_curves = score.getPresence(file, first, last, aggreg);

    /* Compute grid ranges */
    int twidth = fonts[FONT_NORMAL]->getTextWidth(_("Turn"));
    int mwidth = fonts[FONT_NORMAL]->getTextWidth(numToString(max_score));

    left_margin   = (twidth > mwidth ? twidth : mwidth) + 2;
    right_margin  = twidth + 2;
    bottom_margin = font_heights[FONT_NORMAL];

    /* Compute auxiliary grid position */
    /* First, figure out whether there's decay: */
    aux_decay = 0;
    if (score.getNumParts() == 1 && file.getScoreId(score.getPart(0)) == ScoreId_BuildPoints && host.isPHost()) {
        aux_decay = config.PALDecayPerTurn(::getPlayerId());
        if (aux_decay < 0 || aux_decay >= 100)
            aux_decay = 0;
    }

    if (aux_decay == 0) {
        /* No decay, so we'll be making regular lines. We wish the first line to be a little
           below the maximum, and we want no more than 5 grid lines. We try all distances
           that have nice numbers, i.e. 1,2,5,10,20,50,100,etc. */
        static const int nice_numbers[] = { 1, 2, 5 };
        int32 limit = (max_score * 23)/25;
        unsigned int nice_index = 0;
        int32 nice_scale = 1;
        while ((aux_scale = nice_numbers[nice_index] * nice_scale) * 5 < limit) {
            ++nice_index;
            if (nice_index >= countof(nice_numbers)) {
                nice_index = 0;
                nice_scale *= 10;
            }
        }
        aux_limit = limit;
    } else {
        /* There ist decay. Let the first line start above the maximum, so that it just
           comes in range with the first step. Since we'll not be labelling the lines,
           numbers don't have to be nice. */
        aux_limit = max_score * 100 / (100-aux_decay);
        aux_scale = aux_limit / 5;
        if (aux_scale < 10)
            aux_scale = 10;
    }
}

/** Draw chart for one player.
    \param can Canvas to draw on
    \param player Player whose chart to draw */
void
WScoreChart::drawPlayerChart(GfxContext& ctx, int player)
{
    int prev_x = 0, prev_y = 0;
    int first_turn = file.getRecordByIndex(first).getTurnNumber();
    int turns = file.getRecordByIndex(last-1).getTurnNumber() - first_turn;

    int width  = extent.w - right_margin - left_margin - 2;
    int height = extent.h - bottom_margin - 1;
    int x0     = extent.x + left_margin;
    int y0     = extent.y + height + 1;

    bool drawing = false;

    ctx.setColor(WRaceList::colors[player-1]);

    for (uint32 i = first; i < last; ++i) {
        const GStatRecord& r = file.getRecordByIndex(i);

        int32 value = score.get(r, player, aggreg);
        if (value >= 0) {
            int x = x0 + width*(r.getTurnNumber()-first_turn) / turns;
            int y = y0 - height*value / max_score;
            if (drawing)
                drawLine(ctx, GfxPoint(prev_x, prev_y), GfxPoint(x, y));
            drawPixel(ctx, GfxPoint(x-1, y));
            drawPixel(ctx, GfxPoint(x+1, y));
            drawPixel(ctx, GfxPoint(x, y-1));
            drawPixel(ctx, GfxPoint(x, y+1));
            prev_x = x;
            prev_y = y;
            drawing = true;
        } else {
            drawing = false;
        }
    }
}

/** Find victory condition. */
void
WScoreChart::findVictoryCondition()
{
    victory_level = -1;
    victory_turns_needed = 0;
    victory_turns.setAll(0);

    /* Only single-part charts can have a sensible victory condition. */
    if (score.getNumParts() != 1 || score.getFactor(0) != 1)
        return;

    /* Team scores cannot have victory condition. */
    if (aggreg != ScoreByPlayer)
        return;

    /* Do we have a description for this score? */
    const TScoreDescription* pdesc = file.getDescriptionForScoreId(file.getScoreId(score.getPart(0)));
    if (pdesc == 0)
        return;

    /* Does this description specify a victory condition? */
    if (pdesc->win_limit <= 0)
        return;

    victory_level = pdesc->win_limit;
    victory_turns_needed = pdesc->turn_limit > 0 ? pdesc->turn_limit : 1;

    /* Find victory turns, for each player. */
    for (uint32 index = file.getNumTurns(); index > 0; --index) {
        const GStatRecord& record = file.getRecordByIndex(index-1);
        bool any_player_over_limit = false;
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            if (score.get(record, i, aggreg) >= victory_level) {
                victory_turns[i] = record.getTurnNumber();
                any_player_over_limit = true;
            }
        }
        if (!any_player_over_limit)
            break;
    }
}

/** Show tooltip at mouse position \c pt.
    \todo Here is lot of room for improvement. Instead of just a tooltip, PCC 1.x
    displays kind-of a crosshair to make it obvious that the tooltip describes an X/Y,
    not a specific data point. On the other hand, it would make sense to "lock" onto
    data points; the tooltip would then contain, say, a second line saying whose data
    point we look at. */
void
WScoreChart::doTooltip(GfxPoint pt)
{
    // Figure out coordinates, see drawPlayerChart
    int first_turn = file.getRecordByIndex(first).getTurnNumber();
    int turns = file.getRecordByIndex(last-1).getTurnNumber() - first_turn;

    int width  = extent.w - right_margin - left_margin - 2;
    int height = extent.h - bottom_margin - 1;
    int x0     = extent.x + left_margin;
    int y0     = extent.y + height + 1;

    // Wield out invalid values quickly
    if (pt.x < x0 || pt.y > y0)
        return;

    // Reverse coordinate transform with rounding
    int turn  = ((pt.x - x0) * turns     + width/2)  / width + first_turn;
    int value = ((y0 - pt.y) * max_score + height/2) / height;
    if (turn < first_turn || turn > first_turn + turns || value < 0 || value > max_score)
        return;

    // Forward coordinate transform, see drawPlayerChart
    int x = x0 + width*(turn-first_turn) / turns;
    int y = y0 - height*value / max_score;

    ::doTooltip(format(_("Turn %d: %d"), turn, value), GfxPoint(x, y));
}

bool
WScoreChart::handleEvent(const UIEvent& e, bool second_pass)
{
    if (tt_ctl.checkEvent(e))
        doTooltip(tt_ctl.getTooltipPosition());

    if (e.type == ev_Key) {
        if (e.key.code == ss_Ctrl + SDLK_LEFT) {
            setSelectedWidget();
            int n = current_player - 1;
            if (n <= 0)
                n = NUM_PLAYERS;
            setCurrentPlayer(n);
            return true;
        } else if (e.key.code == ss_Ctrl + SDLK_RIGHT) {
            setSelectedWidget();
            int n = current_player + 1;
            if (n > NUM_PLAYERS)
                n = 1;
            setCurrentPlayer(n);
            return true;
        } else {
            return UIWidget::handleEvent(e, second_pass);
        }
    } else {
        return UIWidget::handleEvent(e, second_pass);
    }
}

void
WScoreChart::onStateChange(int state, bool enable)
{
    tt_ctl.checkState();
    UIWidget::onStateChange(state, enable);
}
