/**
  *  \file game/fcode-load.cc
  *  \brief Friendly Code Loading
  *
  *  This module provides functions for loading friendly code definitions
  *  from definition files (fcodes.cc).
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/fcode.h"
#include "util/console.h"
#include "cpluslib/string.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "util/except.h"
#include "io/streams.h"
#include "io/textfile.h"
#include "util/modinit.h"
#include "game/gamedirs.h"

/** Load friendly code list from a file. Syntax errors are logged on
    the console.
    \param inp   the file */
void
GFCodeList::loadFromFile(Stream& inp)
{
    TextFile tf(inp);
    string_t line;
    while (tf.getLine(line)) {
        line = strTrim(line);
        if (line.empty() || line[0] == ';')
            continue;

        string_t::size_type p = line.find_first_of("=,");
        if (p == string_t::npos) {
            console.write(LOG_ERROR, format(_("%s:%d: missing delimiter"), inp.getName(), tf.getLineNr()));
        } else if (line[p] == ',') {
            string_t fc = strTrim(line.substr(0, p));
            line.erase(0, p+1);
            if (fc.length() > 3)
                console.write(LOG_ERROR, format(_("%s:%d: friendly code too long; truncated"), inp.getName(), tf.getLineNr()));
            try {
                addFCode(GFCode(fc, line));
            }
            catch(UserException& e) {
                console.write(LOG_ERROR, format("%s:%d: %s", inp.getName(), tf.getLineNr(), e.what()));
            }
        } else {
            /* FIXME: additional assignments. These still need to be
               implemented. */
        }
    }
    sort();
}

/** Load friendly codes. This initializes all the friendly-code
    related data structures. */
static void
initFCodes()
{
    GFCodeList::getInstance().clear();
    Ptr<Stream> s = spec_file_dir->openFileNT("fcodes.cc", Stream::C_READ);
    if (s) {
        GFCodeList::getInstance().loadFromFile(*s);
    }
    s = spec_file_dir->openFileNT("fcodes.usr", Stream::C_READ);
    if (s) {
        GFCodeList::getInstance().loadFromFile(*s);
    }

    GFCode::clearExtraFC();
    s = spec_file_dir->openFileNT("xtrfcode.txt", Stream::C_READ);
    if (s) {
        GFCode::loadExtraFC(*s);
    }
}

const ModuleInitializer mod_fcode[] = {
    { OnEnterNoGame | OnEnterDirectory, initFCodes },
    { 0, 0 },
};
