/**
  *  \file gfx/font.cc
  *  \brief Font Drivers
  *
  *  This module contains the basic font driver interface, and the
  *  implementation of a standard driver.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/font.h"
#include "io/io.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/assert.h"
#include "cpluslib/utf8.h"

/************************ "Complicated" Functions ************************/

/** Output Text, using Alignment Parameters.
    \param ctx    Output
    \param x,y    Anchor point
    \param text   Text to output */
void
outText(GfxContext& ctx, int x, int y, string_t text)
{
    GfxFont& fnt   = ctx.getFont();
    GfxPoint align = ctx.getTextAlign();
    if (align.x)
        x -= fnt.getTextWidth(text) * align.x / 2;
    if (align.y)
        y -= fnt.getTextHeight(text) * align.y / 2;
    fnt.outText(ctx, GfxPoint(x, y), text);
}

/** Output Text with fixed maximum width. Outputs the text so that it
    occupies at most maxw pixels on the screen. When non-transparent output
    is used, draws appropriate boxes so that exactly the specified area is
    covered. Otherwise like outText().

    This function moves the graphics cursor, depending on the text
    justification:
    - when using left-justified text, the cursor is set to just after
      the output text;
    - when using right-justified text, the cursor is set to just before
      the output text;
    - when using centered text, the cursor is set to (x,y)

    The rationale is that (when not using centered output) one can do
    several outText()s in a row and always know the correct place without
    explicitly having to compute the width. */
void
outTextF(GfxContext& ctx, int x, int y, int maxw, string_t text)
{
    int width, height;
    GfxFont& fnt = ctx.getFont();

    /* Limit text width */
    while (1) {
        width = fnt.getTextWidth(text);
        if (width <= maxw || text.empty())
            break;

        /* We must remove some characters. Make sure to not hit a
           UTF-8 rune in the middle. */
        string_t::size_type n = text.length() - 1;
        while (n > 0 && Utf8Reader::isContinuationByte(text[n]))
            --n;
        text.erase(n);
    }

    height = fnt.getTextHeight(text);
    GfxPoint align = ctx.getTextAlign();
    int newcy = y;
    y -= height * align.y / 2;

    switch (align.x) {
     case 0:
        /* Left */
        if (!ctx.isTransparentBackground())
            drawBackground(ctx, GfxRect(x, y, maxw, height));
        fnt.outText(ctx, GfxPoint(x, y), text);
        ctx.setCursor(GfxPoint(x + width, newcy));
        break;
     case 1:
        /* Centered */
        {
            int xl = x - maxw/2;
            int xtl = x - width/2;
            if (!ctx.isTransparentBackground())
                drawBackground(ctx, GfxRect(xl, y, maxw, height));
            fnt.outText(ctx, GfxPoint(xtl, y), text);
            ctx.setCursor(GfxPoint(x, newcy));
            break;
        }
     case 2:
        /* Right */
        if (!ctx.isTransparentBackground())
            drawBackground(ctx, GfxRect(x - maxw, y, maxw, height));
        fnt.outText(ctx, GfxPoint(x - width, y), text);
        ctx.setCursor(GfxPoint(x - width, newcy));
        break;
     default:
        ASSERT(0);
    }
}
