/**
  *  \file int/if/msgif.cc
  *  \brief Interpreter Game Interface: Messages
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/msgif.h"
#include "game/msg.h"
#include "int/if/ifutil.h"
#include "int/compound.h"
#include "game/global.h"
#include "cpluslib/misc.h"
#include "int/interror.h"
#include "io/textfile.h"
#include "cpluslib/format.h"
#include "int/file.h"

namespace {
    enum {
        impId,
        impLines,
        impGroup,
        impKilled,
        impText,
        impFullText,

        immWrite,
    };
    enum { MessageProperty };

    /** Message "Text()" Property. */
    class IntMessageText : public IntIndexableValue {
        Ptr<GMessageLines> lines;
     public:
        IntMessageText(Ptr<GMessageLines> lines);
        ~IntMessageText();

        // IntIndexableValue:
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& args, IntValue* value);
        IntContext* makeFirstContext();

        // IntValue:
        IntMessageText* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
    };

    /** Message "Write" Method. */
    class IntMessageWriteProcedure : public IntCallableValue {
        int turn;
        int nr;
        Ptr<GMessageLines> lines;
     public:
        IntMessageWriteProcedure(int turn, int nr, Ptr<GMessageLines> lines);
        ~IntMessageWriteProcedure();

        // IntCallableValue:
        IntValue* call(IntExecutionContext& exc, IntArgBlock& args);
        bool isProcedureCall();

        // IntValue:
        IntMessageWriteProcedure* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
    };
}

/** Implementation of "InMsg(n)". */
IntValue*
IFInmsgGet(IntArgBlock& args)
{
    int32_t n;
    args.checkArgumentCount(1);
    if (!haveDisplayedTurn())
        return 0;

    GMessagebox& inbox = getDisplayedTurn().getInbox();
    if (!checkIntArg(n, args.getNext(), 1, inbox.getCount()))
        return 0;

    return new IntMessageContext(inbox, n-1);
}

/** Implementation of "ForEach InMsg". */
IntContext*
IFInmsgMake()
{
    if (!haveDisplayedTurn())
        return 0;

    GMessagebox& inbox = getDisplayedTurn().getInbox();
    if (!inbox.getCount())
        return 0;

    return new IntMessageContext(inbox, 0);
}

/*************************** IntMessageContext ***************************/

/** Constructor.
    \param mbox Message box to observe
    \param nr Message number */
IntMessageContext::IntMessageContext(GMessagebox& mbox, int nr)
    : mbox(mbox),
      nr(nr),
      line_cache(0)
{ }

IntMessageContext::~IntMessageContext()
{ }

void
IntMessageContext::createLineCache()
{
    if (!line_cache) {
        line_cache = new GMessageLines();
        splitMessage(*line_cache, mbox.getText(nr));
    }
}

static const TIntMapping msg_mapping[] = {
    { "FULLTEXT", impFullText, MessageProperty },
    { "GROUP",    impGroup,    MessageProperty },
    { "ID",       impId,       MessageProperty },
    { "KILLED",   impKilled,   MessageProperty },
    { "LINES",    impLines,    MessageProperty },
    { "TEXT",     impText,     MessageProperty },
    { "WRITE",    immWrite,    MessageProperty },
};

// IntContext:
int32_t
IntMessageContext::lookup(const string_t& name)
{
    return lookupName(name, msg_mapping, countof(msg_mapping));
}

void
IntMessageContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntMessageContext::get(int32_t index)
{
    switch (msg_mapping[index].index) {
     case impId:
        return makeIntValue(nr+1);
     case impLines:
        createLineCache();
        return makeIntValue(line_cache->size());
     case impGroup:
        return makeStringValue(mbox.getHeading(nr));
     case impKilled:
        return makeBoolValue(::getMessageFilter().isFiltered(mbox.getHeading(nr)));
     case impText:
        createLineCache();
        return new IntMessageText(line_cache);
     case impFullText:
        return makeStringValue(mbox.getText(nr));

     case immWrite:
        createLineCache();
        return new IntMessageWriteProcedure(getDisplayedTurn().getTurnNumber(), nr, line_cache);
     default:
        return 0;
    }
}

bool
IntMessageContext::next()
{
    if (nr+1 < mbox.getCount()) {
        line_cache = 0;
        ++nr;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntMessageContext::clone()
{
    return new IntMessageContext(mbox, nr);
}

GObject*
IntMessageContext::getObject()
{
    return 0;
}

// IntValue:
string_t
IntMessageContext::toString(bool /*readable*/)
{
    // FIXME: we can do better
    return "#<message>";
}

void
IntMessageContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

/***************************** IntMessageText ****************************/

IntMessageText::IntMessageText(Ptr<GMessageLines> lines)
    : lines(lines)
{ }

IntMessageText::~IntMessageText()
{ }

// IntIndexableValue:
IntValue*
IntMessageText::get(IntArgBlock& args)
{
    int32_t n;
    args.checkArgumentCount(1);
    if (!checkIntArg(n, args.getNext(), 1, lines->size()))
        return 0;
    return makeStringValue((*lines)[n-1]);
}

void
IntMessageText::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntContext*
IntMessageText::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
IntMessageText*
IntMessageText::clone()
{
    return new IntMessageText(lines);
}

string_t
IntMessageText::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntMessageText::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

/************************ IntMessageWriteProcedure ***********************/

IntMessageWriteProcedure::IntMessageWriteProcedure(int turn, int nr, Ptr<GMessageLines> lines)
    : turn(turn),
      nr(nr),
      lines(lines)
{ }

IntMessageWriteProcedure::~IntMessageWriteProcedure()
{ }

// IntCallableValue:
IntValue*
IntMessageWriteProcedure::call(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    TextFile* tf;
    int32_t flags = 0;
    args.checkArgumentCount(1, 2);
    if (!checkFileArg(tf, args.getNext()))
        return 0;
    checkFlagArg(flags, 0, args.getNext(), "R");

    if (flags == 0) {
        tf->writeLine(format("=== Turn %d ===", turn));
        tf->writeLine(format("--- Message %d ---", nr+1));
    }
    for (GMessageLines::iterator i = lines->begin(); i != lines->end(); ++i) {
        tf->writeLine(*i);
    }
    return 0;
}

bool
IntMessageWriteProcedure::isProcedureCall()
{
    return true;
}

// IntValue:
IntMessageWriteProcedure*
IntMessageWriteProcedure::clone()
{
    return new IntMessageWriteProcedure(turn, nr, lines);
}

string_t
IntMessageWriteProcedure::toString(bool /*readable*/)
{
    return "#<procedure>";
}

void
IntMessageWriteProcedure::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}
