/**
  *  \file int/selexpr.cc
  *  \brief Interpreter: Compilation of Selection Expressions
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/selexpr.h"
#include "game/sel.h"
#include "int/lex.h"
#include "int/interror.h"

/** Compile optional type-mask in set expression. A parenthesized expression
    after "S" or "P" means to select all ships or planets from that expression. */
static void
compileOptionalTypeMask(IntTokenizer& tok, string_t& expr)
{
    if (tok.checkAdvance(tok.tLParen)) {
        compileSelectionExpression(tok, expr);
        expr += char(GMultiSelection::opAnd);
        if (!tok.checkAdvance(tok.tRParen))
            throw IntError::expectSymbol(")");
    }
}

/** Compile "factor" for set expression.

    factor ::= "not" factor
             | "-" factor                 // this production not in PCC 1.x
             | "S" type-mask_opt
             | "SHIPS" type-mask_opt
             | "P" type-mask_opt
             | "PLANETS" type-mask_opt
             | "CURRENT"
             | "(" expression ")"
             | "A" | "B" | "C" | "D" | "E" | "F" | "G" | "H"
             | "0" | "1" */
static void
compileFactor(IntTokenizer& tok, string_t& expr)
{
    // search.pas::CompileFactor
    if (tok.checkAdvance(tok.tNOT) || tok.checkAdvance(tok.tMinus)) {
        compileFactor(tok, expr);
        expr += char(GMultiSelection::opNot);
    } else if (tok.checkAdvance("S") || tok.checkAdvance("SHIPS")) {
        expr += char(GMultiSelection::opShip);
        compileOptionalTypeMask(tok, expr);
    } else if (tok.checkAdvance("P") || tok.checkAdvance("PLANETS")) {
        expr += char(GMultiSelection::opPlanet);
        compileOptionalTypeMask(tok, expr);
    } else if (tok.checkAdvance("CURRENT")) {
        expr += char(GMultiSelection::opCurrent);
    } else if (tok.checkAdvance(tok.tLParen)) {
        compileSelectionExpression(tok, expr);
        if (!tok.checkAdvance(tok.tRParen))
            throw IntError::expectSymbol(")");
    } else {
        if (tok.getCurrentToken() == tok.tIdentifier) {
            string_t s = tok.getCurrentString();
            if (s.size() == 1 && s[0] >= 'A' && s[0] < 'A' + GMultiSelection::NUM_SELECTION_LAYERS) {
                expr += char(s[0] - 'A' + GMultiSelection::opFirstLayer);
                tok.readNext();
            } else {
                throw IntError::unknownIdentifier(s);
            }
        } else if (tok.getCurrentToken() == tok.tInteger) {
            if (tok.getCurrentInteger() == 0) {
                expr += char(GMultiSelection::opZero);
                tok.readNext();
            } else if (tok.getCurrentInteger() == 1) {
                expr += char(GMultiSelection::opOne);
                tok.readNext();
            } else {
                throw IntError("Invalid operand in set expression");
            }
        } else if (tok.getCurrentToken() == tok.tEnd) {
            throw IntError("Expecting operand in set expression");
        } else {
            throw IntError("Invalid operand in set expression");
        }
    } 
}

/** Compile "summand" for set expression.

    summand := factor
             | summand "And" factor
             | summand "*" factor */
static void
compileSummand(IntTokenizer& tok, string_t& expr)
{
    compileFactor(tok, expr);
    while (1) {
        if (tok.checkAdvance(tok.tAND) || tok.checkAdvance(tok.tMultiply)) {
            compileFactor(tok, expr);
            expr += char(GMultiSelection::opAnd);
        } else {
            break;
        }
    }
}

/** Compile selection expression. Selection expressions are compiled
    into a simple RPN string, see GMultiSelection.

    \param tok [in] Tokenizer to read from
    \param expr [out] Code will be appended here
    
    expression ::= summand
                 | expression ("+"|"Or") summand
                 | expression "Xor" summand         // this production not in PCC 1.x
                 | expression "-" summand */
void
compileSelectionExpression(IntTokenizer& tok, string_t& expr)
{
    compileSummand(tok, expr);
    while (1) {
        if (tok.checkAdvance(tok.tOR) || tok.checkAdvance(tok.tPlus)) {
            compileSummand(tok, expr);
            expr += char(GMultiSelection::opOr);
        } else if (tok.checkAdvance(tok.tXOR)) {
            compileSummand(tok, expr);
            expr += char(GMultiSelection::opXor);
        } else if (tok.checkAdvance(tok.tMinus)) {
            compileSummand(tok, expr);
            expr += char(GMultiSelection::opNot);
            expr += char(GMultiSelection::opAnd);
        } else {
            break;
        }
    }
}
