/**
  *  \file ui/widgets/cardtab.cc
  *  \brief UICardGroupTab widget
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/cardtab.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "ui/skin.h"
#include "ui/widgets/cardgroup.h"
#include "ui/uiglobal.h"


//! Information about a tab. \internal
struct UICardGroupTab::TabInfo {
    int         id;             ///< Widget id of tab.
    string_t    name;           ///< Name displayed on top of screen.
    UIKey       key;            ///< Hot-key.

    TabInfo(int id, const string_t name, UIKey key);
    ~TabInfo();
};

UICardGroupTab::TabInfo::TabInfo(int id, const string_t name, UIKey key)
    : id(id), name(name), key(key)
{ }

UICardGroupTab::TabInfo::~TabInfo()
{ }

/** Constructor.
    \param group  the UICardGroupTab we want to control. */
UICardGroupTab::UICardGroupTab(UICardGroup& group)
    : group(group),
      group_conn(observe(group.sig_focus_changed))
{ }

UICardGroupTab::~UICardGroupTab()
{
    group_conn.disconnect();
}

void
UICardGroupTab::drawContent(GfxCanvas& can)
{
    int x = getExtent().x;
    int top = getExtent().y;
    int bot = getExtent().y + getExtent().h;

    // figure out current selected item
    int cur_id = getCurrentTab();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(group.getSkin()); // used to draw interior of tabs which is visually inside the group
    ctx.useFont(*fonts[FONT_TITLE]);

    for (std::size_t i = 0; i < tabs.size(); ++i) {
        // first, draw lines at bottom
        int this_width = fonts[FONT_TITLE]->getTextWidth(tabs[i]->name) + 20;
        int lwidth = tabs[i]->id == cur_id ? 10 : this_width+16;
        can.drawHLine(GfxPoint(x, bot-2), lwidth, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
        can.drawHLine(GfxPoint(x, bot-1), lwidth, standard_colors[COLOR_GRAY],  SOLID_LINE, GFX_ALPHA_OPAQUE);
        if (tabs[i]->id == cur_id)
            drawBackground(ctx, GfxRect(x + 10, bot - 2, this_width+6, 2));

        // left side of tab
        x += 10;
        can.drawVLine(GfxPoint(x-1, top), bot-2-top, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
        can.drawVLine(GfxPoint(x, top+1), bot-2-top, standard_colors[COLOR_GRAY], SOLID_LINE, GFX_ALPHA_OPAQUE);

        // top
        can.drawHLine(GfxPoint(x, top),   this_width, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
        can.drawHLine(GfxPoint(x, top+1), this_width, standard_colors[COLOR_GRAY], SOLID_LINE, GFX_ALPHA_OPAQUE);

        // content and rhs
        int cx = 0, he = bot - top - 2, lastl = 1, bx = x + this_width;
        for (int l = 1; l < he; ++l) {
            int w = 6 * l / he;
            if (w != cx) {
                drawSolidBar  (can, GfxRect(bx + cx, top + lastl, 2,  l - lastl), standard_colors[COLOR_BLACK]);
                if (lastl == 1 && l != 1)  // HACK HACK HACK
                    ++lastl;
                drawBackground(ctx, GfxRect(x+1,     top + lastl, cx+this_width-1, l - lastl));
                lastl = l;
                cx = w;
            }
        }
        drawBackground(ctx, GfxRect(x+1,     top + lastl, cx+this_width-1, he - lastl));
        drawSolidBar  (can, GfxRect(bx + cx, top + lastl, 2,  he - lastl), standard_colors[COLOR_BLACK]);

        ctx.setColor(tabs[i]->id == cur_id ? UIColor::tc_Heading : UIColor::tc_Static);
        outText(ctx, x + 10, top + 1, tabs[i]->name);

        x += this_width+6;
    }

    int right = getExtent().x + getExtent().w;
    if (x < right) {
        can.drawHLine(GfxPoint(x, bot-2), right-x, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
        can.drawHLine(GfxPoint(x, bot-1), right-x, standard_colors[COLOR_GRAY],  SOLID_LINE, GFX_ALPHA_OPAQUE);
    }
    can.update(getExtent());
}

void
UICardGroupTab::getLayoutInfo(LayoutInfo& li)
{
    int min_x = 0;
    for (std::size_t i = 0; i < tabs.size(); ++i)
        min_x += fonts[FONT_TITLE]->getTextWidth(tabs[i]->name) + 36;

    li.flags = li_FlexibleH;
    li.pref_size = li.min_size = GfxPoint(min_x, font_heights[FONT_TITLE] + 3);
}

bool
UICardGroupTab::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        /* FIXME: I'm not certain whether should be here.
             pro: allows all sorts of Id sequences
             con: UIFocusIterator exists and works fine */
        // tab: next tab
        if (e.key.code == SDLK_TAB) {
            int x = getCurrentTabIndex() + 1;
            if (x >= int(tabs.size()))
                x = 0;
            return setCurrentTab(tabs[x]->id);
        }
        // shift-tab: previous tab
        if (e.key.code == ss_Shift + SDLK_TAB) {
            int x = getCurrentTabIndex() - 1;
            if (x < 0)
                x = tabs.size() - 1;
            return setCurrentTab(tabs[x]->id);
        }
        // letters atop the bar
        for (std::size_t i = 0; i < tabs.size(); ++i) {
            if ((e.key.code & ~ss_Alt) == tabs[i]->key)
                return setCurrentTab(tabs[i]->id);
        }
    } else if (e.type == ev_Mouse && isInside(GfxPoint(e.mouse.x, e.mouse.y)) && e.mouse.b) {
        int x = getExtent().x;
        for (std::size_t i = 0; i < tabs.size(); ++i) {
            int tw = fonts[FONT_TITLE]->getTextWidth(tabs[i]->name) + 20;
            x += 10;
            if (x > e.mouse.x)
                break;
            if (e.mouse.x < x + tw)
                return setCurrentTab(tabs[i]->id);
            x += tw+6;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

/** Add a tab.
    \param id     Id of the widget inside the controlled group.
    \param name   Name to display on the tab bar.
    \param key    Hot-key. The UICardGroupTab will accept both the
                  hot-key, as well as Alt+hot-key to select this tab.
                  The hot-key for a tab will only be recognized when
                  the tab in question is not already selected, so the
                  tab selected with 's' can itself use the hot-key 's'. */
void
UICardGroupTab::addTab(int id, const string_t& name, UIKey key)
{
    tabs.push_back(new TabInfo(id, name, key));
    drawWidget();
}

/** Add a tab. \overload */
void
UICardGroupTab::addTab(int id, const UIKeyString& key)
{
    addTab(id, key.getString(), key.getKey());
}

/** Get currently-selected tab Id. This returns the Id of the widget
    currently shown, -1 if none. */
int
UICardGroupTab::getCurrentTab() const
{
    if (UIBaseWidget* p = group.getFocusedChild())
        return p->getId();
    else
        return -1;
}

/** Set currently-selected tab. This brings the widget with the given
    Id to the front.
    \return true iff there was a change, false if not. */
bool
UICardGroupTab::setCurrentTab(int id)
{
    setSelectedWidget();

    UIBaseWidget* w = group.getWidgetById(id);
    if (!w)
        return false;
    if (group.getFocusedChild() && w->hasState(st_Focused))
        return false;
    group.setFocusedWidget(*w);
    return true;
}

/** Get index of currently-selected tab.
    \return index (0..getNumTabs()-1), or -1 if none selected. */
int
UICardGroupTab::getCurrentTabIndex() const
{
    int i = getCurrentTab();
    for (std::size_t x = 0; x < tabs.size(); ++x)
        if (tabs[x]->id == i)
            return x;
    return -1;
}

/** Set current tab by index.
    \param n tab to select (0..getNumTabs()-1, 0 is left-most).
    \see getCurrentTabIndex
    \see setCurrentTab */
bool
UICardGroupTab::setCurrentTabIndex(int n)
{
    return setCurrentTab(tabs[n]->id);
}
